package mbHP;
import java.awt.image.*;
import java.math.BigDecimal;

abstract public class MandComp {
	
	protected MandelbrotCanvas owner;
	protected WritableRaster raster;
	protected int imageWidth, imageHeight;
	protected BigDecimal xmin, xmax, ymin, ymax;
	protected double xmin_d, xmax_d, ymin_d, ymax_d, dx_d, dy_d;
	protected int digits;  // == 0 for normal precision, otherwise number of digits used
	protected int maxIterations;
	protected boolean stretchPalette;
	
	private Thread runner;
	protected volatile boolean running;
	protected static volatile int drawCount;
	
   
	synchronized public void install(MandelbrotCanvas canvas) {
		install(canvas,0);
	}

	synchronized public void install(MandelbrotCanvas canvas, final int delay) {
		if (running)
			cancel();
		drawCount++;
    	owner = canvas;
    	if (canvas == null)
    		return;
    	owner.startedRunning();
    	BufferedImage image = owner.getImage();
    	imageWidth = image.getWidth();
    	imageHeight = image.getHeight();
    	raster = image.getRaster();
    	xmin = owner.getXmin();
    	xmax = owner.getXmax();
    	ymin = owner.getYmin();
    	ymax = owner.getYmax();
		xmin_d = xmin.doubleValue();
		xmax_d = xmax.doubleValue();
		ymin_d = ymin.doubleValue();
		ymax_d = ymax.doubleValue();
		dx_d = (xmax_d - xmin_d)/imageWidth;
		dy_d = (ymax_d - ymin_d)/imageHeight;
    	maxIterations = owner.getMaxIterations();
    	stretchPalette = owner.getStretchPalette();
		if (dx_d > MandelbrotCanvas.HP_CUTOFF && dy_d > MandelbrotCanvas.HP_CUTOFF)
			digits = 0;
		else {
		    digits = xmin.scale();
		    createHPData();
		}
    	if (digits == 0)
    		owner.setMessage("Drawing with Normal Precision, MaxItertions = " + maxIterations);
    	else
    		owner.setMessage("Drawing with High Precision (" + digits + "digits), MaxItertions = " + maxIterations);
    	runner = new Thread() {
    		public void run() {
    			try {
    				int myDrawCount = drawCount;
    				running = true;
    				if (delay > 0) {
    					try {
    						synchronized(MandComp.this) {
    							MandComp.this.wait(delay);
    						}
    					}
    					catch (InterruptedException e) {
    					}
    				}
    				else
    					Thread.yield();
    				try {
    					if (running && drawCount == myDrawCount)
    						MandComp.this.run(myDrawCount);
    				}
    				catch (Exception e) {
    					//System.out.println(e);
    					//e.printStackTrace();
    				}
    			}
    			finally {
    				if (owner != null)
    					owner.doneRunning();
    				synchronized(MandComp.this) {
    					running = false;
    					owner = null;
    					xs = ys = null;
    					work1 = work2 = work3 = zx = zy = null;
    					MandComp.this.notify();
    				}
    			}
    		}
    	};
    	runner.start();
    }
	
	public boolean isRunning() {
		return running;
	}
    
    synchronized public void cancel() {
    	if (! running)
    		return;
    	if (runner != null) {
    		if (runner.isAlive()) {
    			running = false;
    			work1 = zy = null;  //force NullPointerException
    			notify();
    			try {
    				wait(5000);
    			}
    			catch (InterruptedException e) {
    			}
    		}
    		runner = null;
    	}
    }
    
    protected int countIterations(int[] x, int[] y) {
    	System.arraycopy(x,0,zx,0,chunks);
    	System.arraycopy(y,0,zy,0,chunks);
    	int count = 0;
    	//System.out.println("\n\nx = " + convertBack(x,chunks) + ", y = " + convertBack(y,chunks) );
    	while (count < maxIterations) {
    		if ((zx[0] & 0xFFC0) != 0 && (zx[0] & 0xFFC0) != 0xFFC0)
    			break;
    		if ((zy[0] & 0xFFC0) != 0 && (zy[0] & 0xFFC0) != 0xFFC0)
    			break;
    		System.arraycopy(zx,0,work1,0,chunks);
    		add(work1,work1,chunks);
    		multiply(work1,zy,chunks);
    		add(work1,y,chunks);  // work1 is now 2*zx*zy +y, the final next value for zy
    		System.arraycopy(zx,0,work2,0,chunks);
    		add(work2,zy,chunks);  // work2 = zx + zy
    		negate(zy,chunks);  // Done with old zy; let zy = -zy
    		add(zx,zy,chunks);  // zx = zx - (old zy)
    		multiply(zx,work2,chunks); // zx is now old (zx - zy)*(zx + zy) = zx*zx - zy*zy
    		add(zx,x,chunks);  // zx is now zx*zx - zy*zy +x, its correct final value
    		System.arraycopy(work1,0,zy,0,chunks);
    		count++;
        	//System.out.println("  " + count + ". zx = " + convertBack(zx,chunks) + ", zy = " + convertBack(zy,chunks) );
    	}
    	return count;
    }
     
    protected int countIterations(double x, double y) {
 	   int count = 0;
 	   double zx = x;
 	   double zy = y;
 	   while (count < maxIterations
 	               && zx <100 && zx > -100 && zy < 100 && zy > -100) {
 	      double new_zx = zx*zx - zy*zy + x;
 	      zy = 2*zx*zy + y;
 	      zx = new_zx;
 	      count++;
 	   }
 	   return count;    	
     }
     
	public int getIndexFor(int iterations) {
		if (iterations >= maxIterations)
			return 250;
		else if (stretchPalette)
			return (int)(250*((double)iterations)/maxIterations);
		else
			return iterations % 250;
	}
	
	protected void computeColumn(int columnNumber, int[] data){
		if (digits == 0) {
			double x = xmin_d + dx_d*columnNumber;
			for (int i = 0; i < data.length; i++) {
				double y = ymax_d - dy_d*i;
				int iter = countIterations(x,y);
				data[i] = getIndexFor(iter);
			}
		}
		else {
			for (int i = 0; i < data.length; i++) {
				int iter = countIterations(xs[columnNumber],ys[i]);
				data[i] = getIndexFor(iter);
			}
		}
	}
 
	protected void computeBlockColumn(int column, int blockSize, int[] data) {
		if (digits == 0) {
			double x = xmin_d + dx_d*column;
			for (int row = blockSize/2; row-blockSize/2 < imageHeight; row += blockSize) {
				double y = ymax_d - row*dy_d;
				int iter = countIterations(x,y);
				int color = getIndexFor(iter);
				for (int i = row-blockSize/2; i < row+blockSize/2 && i < imageHeight; i++)
					data[i] = color;
			}
		}
		else {
			for (int row = blockSize/2; row-blockSize/2 < imageHeight; row += blockSize) {
				int iter;
				if (row < imageHeight)
					iter = countIterations(xs[column],ys[row]);
				else
					iter = countIterations(xs[column],ys[imageHeight-1]);
				int color = getIndexFor(iter);
				for (int i = row-blockSize/2; i < row+blockSize/2 && i < imageHeight; i++)
					data[i] = color;
			}
		}
	}
	
	abstract public void run(int myDrawCount);
	
	// ------- support for high-precision calculation ------------
	
	private int[][] xs;
	private int[][] ys;
	
	private int[] work1,work2,work3;
    private int[] zx, zy;
    
    private int chunks;
    
    private BigDecimal twoTo16 = new BigDecimal(0x10000);
    
    private final static double log2of10 = Math.log(10)/Math.log(2);
    
    private void createHPData() {
    	chunks =(int)(digits * log2of10)/16 + 2;
    	xs = new int[imageWidth][];
    	ys = new int[imageHeight][];
    	int[] dx = new int[chunks+1];
    	int[] dy = new int[chunks+1];
    	BigDecimal DX = xmax.subtract(xmin).divide(new BigDecimal(imageWidth),BigDecimal.ROUND_HALF_EVEN);
    	BigDecimal DY = ymax.subtract(ymin).divide(new BigDecimal(imageHeight),BigDecimal.ROUND_HALF_EVEN);
    	convert(dx,DX,chunks+1);
    	convert(dy,DY,chunks+1);
    	xs[0] = new int[chunks+1];
    	convert(xs[0],xmin,chunks+1);
    	for (int i = 1; i < imageWidth; i++) {
    		xs[i] = (int[])xs[i-1].clone();
    		add(xs[i],dx,chunks+1);
    	}
    	ys[imageHeight-1] = new int[chunks+1];
    	convert(ys[imageHeight-1],ymin,chunks+1);
    	for (int i = imageHeight - 2; i >= 0; i--) {
    		ys[i] = (int[])ys[i+1].clone();
    		add(ys[i],dy,chunks+1);
    	}
    	zx = new int[chunks];
    	zy = new int[chunks];
    	work1 = new int[chunks];
       	work2 = new int[chunks];
       	work3 = new int[chunks];
    }
    
    private void convert(int[] x, BigDecimal X, int count) {
    	boolean neg = false;
    	if (X.signum() == -1) {
    		neg = true;
    		X = X.negate();
    	}
    	x[0] = X.intValue();
    	for (int i = 1; i < count; i++) {
    		X = X.subtract(new BigDecimal(x[i-1]));
    		X = X.multiply(twoTo16);
    		x[i] = X.intValue();
    	}
    	if (neg)
    		negate(x,count);
    }
    
    private void add(int[] x, int[] dx, int count) {
    	int carry = 0;
    	for (int i = count - 1; i >= 0; i--) {
    		x[i] += dx[i];
    		x[i] += carry;
    		carry = x[i] >>> 16;
    		x[i] &= 0xFFFF;
    	}
    }
    
    private void multiply(int[] x, int[] y, int count){  // Can't allow x == y !
    	boolean neg1 = (x[0] & 0x8000) != 0;
    	if (neg1)
    		negate(x,count);
    	boolean neg2 = (y[0] & 0x8000) != 0;
    	if (neg2)
    		negate(y,count);
    	if (x[0] == 0) {
    		for (int i = 0; i < count; i++)
    			work3[i] = 0;
    	}
    	else {
    		int carry = 0;
    		for (int i = count-1; i >= 0; i--) {
    			work3[i] = x[0]*y[i] + carry;
    			carry = work3[i] >>> 16;
    			work3[i] &= 0xFFFF;
    		}
    	}
    	for (int j = 1; j < count; j++) {
    		int i = count - j;
    		int carry = (x[j]*y[i]) >>> 16;
    		i--;
    		int k = count - 1;
    		while (i >= 0) {
    			work3[k] += x[j]*y[i] + carry;
    			carry = work3[k] >>> 16;
    			work3[k] &= 0xFFFF;
    			i--;
    			k--;
    		}
    		while (carry != 0 && k >= 0) {
    			work3[k] += carry;
    			carry = work3[k] >>> 16;
    			work3[k] &= 0xFFFF;
    			k--;
    		}
    	}
    	System.arraycopy(work3,0,x,0,count);
    	if (neg2)
    		negate(y,count);
    	if (neg1 != neg2)
    		negate(x,count);
    }
    
    private void negate(int[] x, int chunks) {
    	for (int i = 0; i < chunks; i++)
    		x[i] = 0xFFFF-x[i];
    	++x[chunks-1];
    	for (int i = chunks-1; i > 0 && (x[i] & 0x10000) != 0; i--) {
    		x[i] &= 0xFFFF;
    		++x[i-1];
    	}
    	x[0] &= 0xFFFF;
    }
    
    private BigDecimal convertBack(int[] x, int count) {
        boolean neg = false;
        if ((x[0] & 0x8000) != 0) {
           negate(x,count);
           neg = true;
        }
        BigDecimal X = new BigDecimal(x[0]).setScale(count*6);
        BigDecimal div = twoTo16;
        for (int i = 1; i < count; i++) {
           X = X.add(new BigDecimal(x[i]).setScale(count*4).divide(div,BigDecimal.ROUND_HALF_EVEN));
           div = div.multiply(twoTo16);
        }
        if (neg) {
           X = X.negate();
           negate(x,count);
        }
        return X;
      }
       
}